"""main application module, typically invoked from ttspod CLI"""
# optional system certificate trust
try:
    import truststore
    truststore.inject_into_ssl()
except ImportError:
    pass

# standard modules
try:
    from argparse import ArgumentParser
    from importlib.resources import files
    from os import isatty, path, getcwd
    from pathlib import Path
    from shutil import copy
    from sys import stdin, stdout, exc_info
    from traceback import format_exc
    from validators import url
except ImportError as e:
    print(
        f'Failed to import required module: {e}\n'
        'You may need to re-execute quickstart.sh.\n'
        'See https://github.com/ajkessel/ttspod/blob/main/README.md for details.')
    exit()

# ttspod modules
from version import __version__
from util import get_character, get_lock, release_lock, upgrade


class App(object):
    """ttspod application"""

    def __init__(self):
        self.clean = None
        self.config_path = None
        self.debug = None
        self.dry = None
        self.engine = None
        self.force = None
        self.generate = None
        self.got_pipe = None
        self.log = None
        self.main = None
        self.model = None
        self.quiet = None
        self.title = None
        self.gpu = None
        self.upgrade = False
        self.wallabag = None
        self.pocket = None
        self.insta = None
        self.url = None

    def parse(self):
        """parse command-line arguments"""
        parser = ArgumentParser(
            description='Convert any content to a podcast feed.')
        parser.add_argument('url', nargs='*', action='store', type=str, default="",
                            help="specify any number of URLs or local documents "
                            "(plain text, HTML, PDF, Word documents, etc) "
                            "to add to your podcast feed")
        parser.add_argument("-c", "--config", nargs='?', const='AUTO', default=None,
                            help="specify path for config file "
                            "(default ~/.config/ttspod.ini if it exists, "
                            "otherwise .env in the current directory)"
                            )
        parser.add_argument("-g", "--generate", nargs='?', const='AUTO', default=None,
                            help="generate a new config file"
                            "(default ~/.config/ttspod.ini if ~/.config exists, "
                            "otherwise .env in the current directory)"
                            )
        parser.add_argument("-w", "--wallabag", nargs='?', const='audio', default=None,
                            help="add unprocessed items with specified tag (default audio) "
                            "from your wallabag feed to your podcast feed")
        parser.add_argument("-i", "--insta", nargs='?', const='audio', default=None,
                            help="add unprocessed items with specified tag (default audio) "
                            "from your instapaper feed to your podcast feed, "
                            "or use tag ALL for default inbox")
        parser.add_argument("-p", "--pocket", nargs='?', const='audio', default=None,
                            help="add unprocessed items with specified tag (default audio) "
                            "from your pocket feed to your podcast feed")
        parser.add_argument("-l", "--log", nargs='?', const='ttspod.log',
                            default=None, help="log all output to specified filename "
                            "(default ttspod.log)")
        parser.add_argument("-q", "--quiet", nargs='?', default=None,
                            help="no visible output (all output will go to log if specified)")
        parser.add_argument(
            "-d", "--debug", action='store_true', help="include debug output")
        parser.add_argument("-r", "--restart", action='store_true',
                            help="wipe state file clean and start new podcast feed")
        parser.add_argument("-f", "--force", action='store_true',
                            help="force addition of podcast even if "
                            "cache indicates it has already been added")
        parser.add_argument("-t", "--title", action='store',
                            help="specify title for content provided via pipe")
        parser.add_argument("-e", "--engine", action='store',
                            help="specify TTS engine for this session "
                            "(whisper, coqui, openai, eleven)")
        parser.add_argument("-m", "--model", action='store',
                            help="specify model to use with engine "
                            "(for use with Coqui, OpenAI, or Eleven)")
        parser.add_argument("-s", "--sync", action='store_true',
                            help="sync podcast episodes and state file")
        parser.add_argument("-n", "--dry-run", action='store_true',
                            help="do not actually create or sync audio files")
        parser.add_argument("--nogpu", action='store_true',
                            help="disable GPU support (try this if you're having trouble on Mac)")
        parser.add_argument("-u", "--upgrade", action='store_true',
                            help="upgrade to latest version")
        parser.add_argument("-v", "--version", action='store_true',
                            help="print version number")
        args = parser.parse_args()
        self.generate = args.generate
        if self.generate:
            if self.generate == "AUTO":
                self.generate_env_file(None)
            else:
                self.generate_env_file(self.generate)
            return False
        if args.version:
            print(__version__)
            return False
        self.config_path = args.config
        self.debug = args.debug
        self.quiet = args.quiet
        if not self.quiet:
            print(f'TTSPod v{__version__}')
        if self.quiet:
            self.debug = False
        self.log = args.log
        self.dry = args.dry_run
        self.gpu = 0 if args.nogpu else None
        self.force = args.force
        self.clean = args.restart
        self.title = args.title if hasattr(args, 'title') else None
        self.engine = args.engine if hasattr(
            args, 'engine') else None
        self.model = args.model if hasattr(
            args, 'model') else None
        self.got_pipe = not isatty(stdin.fileno())
        self.wallabag = args.wallabag
        self.pocket = args.pocket
        self.insta = args.insta
        self.url = args.url
        if args.upgrade:
            upgrade(force=self.force, debug=self.debug)
            return False
        if not (
            args.url or
            args.wallabag or
            args.pocket or
            args.sync or
            self.got_pipe or
            args.insta
        ):
            parser.print_help()
            return False
        return True

    def generate_env_file(self, env_file):
        """generate a new .env file"""
        if not env_file:
            if path.isdir(path.join(Path.home(), '.config')):
                env_file = path.join(Path.home(), '.config', 'ttspod.ini')
            else:
                env_file = path.join(getcwd(), '.env')
        if path.isdir(env_file):
            env_file = path.join(env_file, '.env')
        if path.isfile(env_file):
            check = False
            while not check:
                stdout.write(
                    f'{env_file} already exists. Do you want to overwrite? (y/n) ')
                stdout.flush()
                check = get_character()
                if isinstance(check, bytes):
                    check = check.decode()
                if not (check == 'y' or check == 'n'):
                    check = False
                elif check == 'n':
                    stdout.write('exiting...\n')
                    exit()
        try:
            copy(files('ttspod').joinpath('data', 'dotenv.env'), env_file)
            print(f'{env_file} written. Now edit to run ttspod.')
        except Exception as err:  # pylint: disable=broad-except
            print(f'Error creating {env_file}: {err}\n.')
        exit()

    def run(self):
        """primary app loop"""
        try:
            if not get_lock():
                if not self.force:
                    print(
                        'Another instance of ttspod was detected running. '
                        'Execute with -f or --force to force execution.')
                    return False
                else:
                    release_lock()
            # this import is slow (loads TTS engines), so only import when needed
            # there is probably a better way to do this by refactoring
            from main import Main  # pylint: disable=import-outside-toplevel
            self.main = Main(
                debug=self.debug,
                config_path=self.config_path,
                engine=self.engine,
                model=self.model,
                force=self.force,
                dry=self.dry,
                clean=self.clean,
                logfile=self.log,
                gpu=self.gpu,
                quiet=self.quiet
            )
            if self.got_pipe:
                pipe_input = str(stdin.read())
                if pipe_input:
                    self.main.process_content(pipe_input, self.title)
            if self.wallabag:
                self.main.process_wallabag(self.wallabag)
            if self.pocket:
                self.main.process_pocket(self.pocket)
            if self.insta:
                self.main.process_insta(self.insta)
            for i in self.url:
                if url(i):
                    self.main.process_link(i, self.title)
                elif path.isfile(path.expanduser(i)):
                    self.main.process_file(path.expanduser(i), self.title)
                else:
                    print(f'command-line argument {i} not recognized')
            return self.main.finalize()
        # pylint: disable=W0718
        # global exception catcher for application loop
        except Exception:
            exc_type, _, exc_tb = exc_info()
            fname = path.split(exc_tb.tb_frame.f_code.co_filename)[1]
            print('Error occurred:\n', exc_type, fname, exc_tb.tb_lineno)
            if self.debug:
                print('-----Full Traceback-----\n', format_exc())
        # pylint: enable=W0718

        finally:
            release_lock()


def main():
    """nominal main loop to read arguments and execute app"""
    app = App()
    if app.parse():   # parse command-line arguments
        # only import remaining modules if we have something to do
        app.run()     # run the main workflow


if __name__ == "__main__":
    main()



"""Xtts generator"""
try:
    import truststore
    truststore.inject_into_ssl()
except ImportError:
    pass
try:
    from os import environ as env
    from platform import processor
    from pprint import pprint
    from warnings import simplefilter
    from TTS.api import TTS
    from transformers import pytorch_utils
    import torch
    import torchaudio
except ImportError as e:
    print(
        f'Failed to import required module: {e}\n'
        'You may need to re-execute quickstart.sh.\n'
        'See https://github.com/ajkessel/ttspod/blob/main/README.md for details.')
    exit()

# from util import patched_isin_mps_friendly
from logger import Logger
simplefilter(action='ignore', category=FutureWarning)

# this attempts to minimize random voice variations
torch.manual_seed(123456789)

# sensible default settings if none are provided
MODEL = 'tts_models--multilingual--multi-dataset--xtts_v2'
TEMPERATURE = 0.2
DEVICE = 'cpu'
VOICE = 'Ana Florence'

if torch.cuda.is_available():
    DEVICE = "cuda"
elif torch.backends.mps.is_available() and processor() != 'i386':
    DEVICE = 'mps'
    env["PYTORCH_ENABLE_MPS_FALLBACK"] = "1"
    # pytorch_utils.isin_mps_friendly = patched_isin_mps_friendly

# cspell: disable
if "cuda" in DEVICE and torch.cuda.get_device_name().endswith("[ZLUDA]"):
    torch.backends.cudnn.enabled = False
    torch.backends.cuda.enable_flash_sdp(False)
    torch.backends.cuda.enable_math_sdp(True)
    torch.backends.cuda.enable_mem_efficient_sdp(False)
# cspell: enable


class Xtts:
    """generator for XttsV2 model"""

    def __init__(self, config=None, log=None, voices=None, gpu='gpu'):
        self.log = log if log else Logger(debug=True)
        self.config = config
        api = TTS(MODEL)
        if voices:
            self.voices = voices
        self.model = api.synthesizer.tts_model
        if gpu == 'cpu':
            self.model.to('cpu')
        else:
            self.model.to(DEVICE)
        self.silence = torch.zeros(int(24000*0.5))
        self.speaker_id = None
        if not voices:
            voices = VOICE
        # TODO: sanity check voice availability with fallback
        if isinstance(voices, list):
            self.gpt_cond_latent, self.speaker_embedding = \
                self.model.get_conditioning_latents(audio_path=self.voices)
        elif isinstance(voices, str):
            self.gpt_cond_latent, self.speaker_embedding = \
                self.model.speaker_manager.speakers[voices].values()
        self.log.write('Xtts generator initialized.',error=False,log_level=2)

    def generate(self, texts=None, output=None):
        """convert a list of texts into an output file"""
        audio_segments = []
        for i, text in enumerate(texts):
            self.log.write(
                f'Processing chunk {i+1} of {len(texts)}:\n{text}',
                error=False,
                log_level=3)
            out = self.model.inference(
                text=text,
                language="en",  # TODO configure or detect language
                gpt_cond_latent=self.gpt_cond_latent,
                speaker_embedding=self.speaker_embedding,
                temperature=TEMPERATURE,
                enable_text_splitting=True
            )
            audio_segments.append(torch.tensor(out["wav"]).unsqueeze(0))
            audio_segments.append(self.silence.unsqueeze(0))
        final_audio = torch.cat(audio_segments, dim=1)
        torchaudio.save(output, final_audio, 24000, format="mp3")
        # TODO sanity check that it actually worked

if __name__ == "__main__":
    xtts = Xtts()
    print("This is the TTSPod XTTS TTS module.")
    pprint(vars(xtts))
    pprint(dir(xtts))



"""Tortoise generator"""
try:
    import truststore
    truststore.inject_into_ssl()
except ImportError:
    pass
try:
    from io import StringIO
    from os import path, environ as env
    from platform import processor
    from pprint import pprint
    from warnings import simplefilter
    from TTS.api import TTS
    from transformers import pytorch_utils
    from contextlib import redirect_stdout, redirect_stderr
    import torch
    import torchaudio
except ImportError as e:
    print(
        f'Failed to import required module: {e}\n'
        'You may need to re-execute quickstart.sh.\n'
        'See https://github.com/ajkessel/ttspod/blob/main/README.md for details.')
    exit()

from util import patched_isin_mps_friendly
from logger import Logger

# suppress spurious FutureWarning from Coqui
simplefilter(action='ignore', category=FutureWarning)

# this attempts to minimize random voice variations
torch.manual_seed(123456789)

# sensible default settings if none are provided
MODEL = 'tts_models/en/multi-dataset/tortoise-v2'
TEMPERATURE = 0.2
DEVICE = 'cpu'
PRESET = 'fast'

if torch.cuda.is_available():
    DEVICE = "cuda"
elif torch.backends.mps.is_available() and processor() != 'i386':
    DEVICE = 'mps'
    env["PYTORCH_ENABLE_MPS_FALLBACK"] = "1"
    pytorch_utils.isin_mps_friendly = patched_isin_mps_friendly

# cspell: disable
if "cuda" in DEVICE and torch.cuda.get_device_name().endswith("[ZLUDA]"):
    torch.backends.cudnn.enabled = False
    torch.backends.cuda.enable_flash_sdp(False)
    torch.backends.cuda.enable_math_sdp(True)
    torch.backends.cuda.enable_mem_efficient_sdp(False)
# cspell: enable


class Tortoise:
    """generator for Tortoise model"""

    def __init__(self, config=None, log=None, voice_dir=None, voice_name=None, gpu='gpu'):
        self.log = log if log else Logger(debug=True)
        self.config = config
        api = TTS(MODEL, progress_bar=False)
        self.model = api.synthesizer.tts_model
        self.tortoise_config = self.model.config
        if gpu == 'cpu':
            self.model.to('cpu')
        else:
            self.model.to(DEVICE)
        self.voice_dir = voice_dir
        self.voice_name = voice_name
        self.silence = torch.zeros(int(24000*0.5)).unsqueeze(0).cpu()
        self.log.write('Tortoise generator initialized.',
                       error=False, log_level=2)

    def generate(self, texts=None, output=None):
        """convert a list of texts into an output file"""
        audio_segments = []
        stdout_buffer = StringIO()
        stderr_buffer = StringIO()
        seed = None
        for i, text in enumerate(texts):
            self.log.write(
                f'Processing chunk {i+1} of {len(texts)}:\n{text}',
                error=False,
                log_level=3)
            try:
                with redirect_stdout(stdout_buffer), redirect_stderr(stderr_buffer):
                    out = self.model.synthesize(
                        text=text,
                        config=self.tortoise_config,
                        speaker_id=self.voice_name,
                        voice_dirs=self.voice_dir,
                        preset=PRESET,
                        use_deterministic_seed=seed,
                        return_deterministic_state=True
                    )
                seed = out['deterministic_seed']
                audio = out['wav'].squeeze(0).cpu()
                audio_segments.append(audio)
                audio_segments.append(self.silence)
            except Exception as e:  # pylint: disable=broad-except
                self.log.write(
                    f'Something went wrong processing {text}: {e}', error=True, log_level=0)
                self.log.write(stdout_buffer.getvalue()+"\n" +
                               stderr_buffer.getvalue(), error=True, log_level=0)
        # TODO: print result in case of failure
        _ = stdout_buffer.getvalue()+"\n"+stderr_buffer.getvalue()
        final_audio = torch.cat(audio_segments, dim=1)
        torchaudio.save(uri=output, src=final_audio,
                        sample_rate=24000, format="mp3")
        if path.isfile(output):
            return output
        else:
            return None


if __name__ == "__main__":
    print("This is the TTSPod Tortoise TTS module. "
          "It is not intended to be run separately except for debugging.")
    tts = Tortoise()
    pprint(vars(tts))
    pprint(dir(tts))

# pylint: disable=line-too-long
    # TEXT = """A Hare was making fun of the Tortoise one day for being so slow.
    # "Do you ever get anywhere?" he asked with a mocking laugh.
    # "Yes," replied the Tortoise, "and I get there sooner than you think. I'll run you a race and prove it."
    # The Hare was much amused at the idea of running a race with the Tortoise, but for the fun of the thing he agreed. So the Fox, who had consented to act as judge, marked the distance and started the runners off.
    # The Hare was soon far out of sight, and to make the Tortoise feel very deeply how ridiculous it was for him to try a race with a Hare, he lay down beside the course to take a nap until the Tortoise should catch up.
    # The Tortoise meanwhile kept going slowly but steadily, and, after a time, passed the place where the Hare was sleeping. But the Hare slept on very peacefully; and when at last he did wake up, the Tortoise was near the goal. The Hare now ran his swiftest, but he could not overtake the Tortoise in time.
    # """
    # tts = Whisper(voice='~/ttspod/working/voices/it')
    # tts.convert(TEXT, "whisper-test.mp3")



"""instapaper input module"""
# optional system certificate trust
try:
    import truststore
    truststore.inject_into_ssl()
except ImportError:
    pass

# only works if instapaper was installed
try:
    import instapaper
    AVAILABLE_INSTAPAPER = True
except ImportError:
    AVAILABLE_INSTAPAPER = False

# TTSPod modules
from logger import Logger
from util import clean_text


class TTSInsta(object):
    """instapaper input"""

    def __init__(self, config, log):
        self.log = log if log else Logger(debug=True)
        self.config = config
        self.p = None
        if not (
            AVAILABLE_INSTAPAPER and
            self.config.username and
            self.config.password and
            self.config.key and
            self.config.secret
        ):
            self.log.write(
                "Instapaper support not enabled, check configuration file.")
            return
        try:
            self.p = instapaper.Instapaper(self.config.key, self.config.secret)
            self.p.login(self.config.username, self.config.password)
        except Exception as err:  # pylint: disable=broad-except
            if 'oauth' in str(err):
                self.log.write('Unable to log in to Instapaper with these credentials:\n'
                               f'Username: {self.config.username}\n'
                               f'Password: {self.config.password}\n'
                               f'API Key: {self.config.key}\n'
                               f'API Secret: {self.config.secret}\n'
                               'Please edit configuration and try again.\n',
                               error=True)
            else:
                self.log.write(f'Instapaper login failed: {err}', error=True)
        return

    def get_items(self, tag):
        """
        retrieve items matching tag

        :param tag: tag name to retrieve or ALL for no filtering
        """
        if not self.p:
            self.log.write("Instapaper support not enabled")
            return []
        bookmarks = self.filter_items(tag)
        if not bookmarks:
            self.log.write(f"No folder or tags found for {tag}")
            return []
        entries = []
        for bookmark in bookmarks:
            self.log.write(
                f'Instapaper content for {bookmark.title} / {bookmark.url}:\n{bookmark.text}',
                log_level=3
            )
            entries.append((bookmark.title, clean_text(
                bookmark.text), bookmark.url))
        return entries

    def filter_items(self, tag):
        """
        search for bookmarks by folder or tag

        :param tag: folder/tag name to retrieve
        """
        if tag == "ALL":
            bookmarks = self.p.bookmarks(limit=1000)
        else:
            bookmarks = self.filter_by_folder(tag)
            bookmarks.extend(self.filter_by_tag(tag))
        return bookmarks

    def filter_by_folder(self, tag):
        """
        search for bookmarks in named folder

        :param tag: folder name to retrieve
        """
        folder_id = []
        folders = self.p.folders()
        folder_id = [folder for folder in folders if folder['title']
                     == tag][0]['folder_id']
        bookmarks = self.p.bookmarks(
            folder=folder_id, limit=1000) if folder_id else []
        return bookmarks

    def filter_by_tag(self, tag):
        """
        search for bookmarks with named tag

        :param tag: tag name to retrieve
        """
        bookmarks = self.p.bookmarks(limit=1000)
        return [bookmark for bookmark in bookmarks if tag in [
            bookmark_tag.get('name') for bookmark_tag in bookmark.tags]]
